#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
@Author  : nickdecodes
@Email   : nickdecodes@163.com
@Usage   :
@FileName: test.py
@DateTime: 2024/1/28 20:13
@SoftWare: 
"""


if __name__ == '__main__':
    pass
